from http import HTTPStatus
from typing import Dict, List

from flask import request
from flask_restx import Resource

import pandas as pd

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE
from mindsdb.api.executor.exceptions import ExecutorException
from mindsdb.api.http.utils import http_error
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.integrations.handlers.web_handler.urlcrawl_helpers import get_all_websites
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.utilities.rag.loaders.file_loader import FileLoader
from mindsdb.integrations.utilities.rag.splitters.file_splitter import FileSplitter, FileSplitterConfig
from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseTable
from mindsdb.utilities import log
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.integrations.utilities.rag.settings import DEFAULT_LLM_MODEL, DEFAULT_RAG_PROMPT_TEMPLATE


from mindsdb_sql.parser.ast import Identifier

logger = log.getLogger(__name__)

_DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON = [
    ("#", "Header 1"),
    ("##", "Header 2"),
    ("###", "Header 3"),
]

_DEFAULT_WEB_CRAWL_LIMIT = 100


def _insert_file_into_knowledge_base(table: KnowledgeBaseTable, file_name: str):
    file_controller = FileController()
    splitter = FileSplitter(FileSplitterConfig())
    file_path = file_controller.get_file_path(file_name)
    loader = FileLoader(file_path)
    split_docs = []
    for doc in loader.lazy_load():
        split_docs += splitter.split_documents([doc])
    doc_objs = []
    for split_doc in split_docs:
        doc_objs.append({
            'content': split_doc.page_content,
        })
    docs_df = pd.DataFrame.from_records(doc_objs)
    # Insert documents into KB
    table.insert(docs_df)


def _insert_web_pages_into_knowledge_base(table: KnowledgeBaseTable, urls: List[str], crawl_depth: int = 1, filters: List[str] = None):
    try:
        # To prevent dependency on langchain_text_splitters unless needed.
        from langchain_text_splitters import MarkdownHeaderTextSplitter
    except ImportError as e:
        logger.error(f'Error importing langchain_text_splitters to insert web page into knowledge base: {e}')
        raise e

    websites_df = get_all_websites(urls, limit=_DEFAULT_WEB_CRAWL_LIMIT, crawl_depth=crawl_depth, filters=filters)
    # Text content is treated as markdown.
    markdown_splitter = MarkdownHeaderTextSplitter(headers_to_split_on=_DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON)

    def append_row_documents(row, all_docs):
        docs = markdown_splitter.split_text(row['text_content'])
        for doc in docs:
            # Link the URL to each web page chunk.
            doc.metadata['url'] = row['url']
        all_docs += docs

    all_docs = []
    websites_df.apply(lambda row: append_row_documents(row, all_docs), axis=1)
    # Convert back to a DF.
    doc_objs = []
    for doc in all_docs:
        doc_objs.append({
            'content': doc.page_content,
            'url': doc.metadata['url']
        })
    docs_df = pd.DataFrame.from_records(doc_objs)
    # Insert documents into KB.
    table.insert(docs_df)


def _insert_select_query_result_into_knowledge_base(query: str, table: KnowledgeBaseTable, project_name: str):
    if not query:
        return
    mysql_proxy = FakeMysqlProxy()
    query_result = mysql_proxy.process_query(query)
    # Use same project as API request for SQL context.
    mysql_proxy.set_context({'db': project_name})
    if query_result.type != SQL_RESPONSE_TYPE.TABLE:
        raise ExecutorException('Query returned no data')

    # Check column name aliases.
    column_names = [c.get('alias', c.get('name')) for c in query_result.columns]
    df_to_insert = pd.DataFrame.from_records(query_result.data, columns=column_names)
    table.insert(df_to_insert)


def _insert_rows_into_knowledge_base(rows: List[Dict], table: KnowledgeBaseTable):
    if not rows:
        return
    df_to_insert = pd.DataFrame.from_records(rows)
    table.insert(df_to_insert)


@ns_conf.route('/<project_name>/knowledge_bases')
class KnowledgeBasesResource(Resource):
    @ns_conf.doc('list_knowledge_bases')
    @api_endpoint_metrics('GET', '/knowledge_bases')
    def get(self, project_name):
        '''List all knowledge bases'''
        session = SessionController()
        project_controller = ProjectController()
        try:
            _ = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        # KB Controller already returns dict.
        return session.kb_controller.list(project_name)

    @ns_conf.doc('create_knowledge_base')
    @api_endpoint_metrics('POST', '/knowledge_bases')
    def post(self, project_name):
        '''Create a knowledge base'''

        # Check for required parameters.
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in POST body'
            )

        knowledge_base = request.json['knowledge_base']
        # Explicitly require embedding model & vector database.
        required_fields = ['name', 'model']
        for field in required_fields:
            if field not in knowledge_base:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing parameter',
                    f'Must provide "{field}" field in "knowledge_base"'
                )
        if 'storage' in knowledge_base:
            if 'database' not in knowledge_base['storage'] or 'table' not in knowledge_base['storage']:
                return http_error(
                    HTTPStatus.BAD_REQUEST,
                    'Missing parameter',
                    'Must provide "database" and "table" field in "storage" param'
                )

        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        kb_name = knowledge_base.get('name')
        existing_kb = session.kb_controller.get(kb_name, project.id)
        if existing_kb is not None:
            # Knowledge Base must not exist.
            return http_error(
                HTTPStatus.CONFLICT,
                'Knowledge Base already exists',
                f'Knowledge Base with name {kb_name} already exists'
            )

        embedding_model_identifier = Identifier(parts=[knowledge_base['model']])
        storage = knowledge_base.get('storage')
        embedding_table_identifier = None
        if storage is not None:
            embedding_table_identifier = Identifier(parts=[storage['database'], storage['table']])
        new_kb = session.kb_controller.add(
            kb_name,
            project.name,
            embedding_model_identifier,
            embedding_table_identifier,
            {}
        )

        return new_kb.as_dict(), HTTPStatus.CREATED


@ns_conf.route('/<project_name>/knowledge_bases/<knowledge_base_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('knowledge_base_name', 'Name of the knowledge_base')
class KnowledgeBaseResource(Resource):
    @ns_conf.doc('get_knowledge_base')
    @api_endpoint_metrics('GET', '/knowledge_bases/knowledge_base')
    def get(self, project_name, knowledge_base_name):
        '''Gets a knowledge base by name'''
        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        existing_kb = session.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )
        return existing_kb.as_dict()

    @ns_conf.doc('update_knowledge_base')
    @api_endpoint_metrics('PUT', '/knowledge_bases/knowledge_base')
    def put(self, project_name: str, knowledge_base_name: str):
        '''Updates a knowledge base.'''
        # Check for required parameters.
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in PUT body'
            )
        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )
        try:
            table = session.kb_controller.get_table(knowledge_base_name, project.id)
        except ValueError:
            # Knowledge Base must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

        kb = request.json['knowledge_base']
        files = kb.get('files', [])
        urls = kb.get('urls', [])
        query = kb.get('query')
        rows = kb.get('rows', [])
        # Optional params for web pages.
        crawl_depth = kb.get('crawl_depth', 1)
        filters = kb.get('filters', [])

        # Load, split, & embed files into Knowledge Base.
        for file_name in files:
            _insert_file_into_knowledge_base(table, file_name)
        # Crawl, split, & embed web pages into Knowledge Base.
        _insert_web_pages_into_knowledge_base(table, urls, crawl_depth=crawl_depth, filters=filters)

        # Execute SELECT statement & embed dataframe into Knowledge Base.
        try:
            _insert_select_query_result_into_knowledge_base(query, table, project_name)
        except ExecutorException as e:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Invalid SELECT query',
                f'Executing "query" failed. Needs to be a valid SELECT statement that returns data: {str(e)}'
            )
        # Convert rows of records into a dataframe & embed into Knowledge Base.
        _insert_rows_into_knowledge_base(rows, table)
        return '', HTTPStatus.OK

    @ns_conf.doc('delete_knowledge_base')
    @api_endpoint_metrics('DELETE', '/knowledge_bases/knowledge_base')
    def delete(self, project_name: str, knowledge_base_name: str):
        '''Deletes a knowledge base.'''
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        session_controller = SessionController()
        existing_kb = session_controller.kb_controller.get(knowledge_base_name, project.id)
        if existing_kb is None:
            # Knowledge Base must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

        session_controller.kb_controller.delete(knowledge_base_name, project_name)
        return '', HTTPStatus.NO_CONTENT


@ns_conf.route('/<project_name>/knowledge_bases/<knowledge_base_name>/completions')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('knowledge_base_name', 'Name of the knowledge_base')
class KnowledgeBaseCompletions(Resource):
    @ns_conf.doc('knowledge_base_completions')
    @api_endpoint_metrics('POST', '/knowledge_bases/knowledge_base/completions')
    def post(self, project_name, knowledge_base_name):
        """
        Add support for LLM generation on the response from knowledge base
        """
        # Check for required parameters.
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in POST body'
            )

        # Check for required parameters
        query = request.json.get('query')
        if query is None:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "query" parameter in POST body'
            )

            logger.error('Missing parameter "query" in POST body')

        llm_model = request.json.get('llm_model')
        if llm_model is None:
            logger.warn(f'Missing parameter "llm_model" in POST body, using default llm_model {DEFAULT_LLM_MODEL}')

        prompt_template = request.json.get('prompt_template')
        if prompt_template is None:
            logger.warn(f'Missing parameter "prompt_template" in POST body, using default prompt template {DEFAULT_RAG_PROMPT_TEMPLATE}')

        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except EntityNotExistsError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )
            logger.error("Project not found, please check the project name exists")

        # Check if knowledge base exists
        table = session.kb_controller.get_table(knowledge_base_name, project.id)
        if table is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

            logger.error("Knowledge Base not found, please check the knowledge base name exists")

        # Get retrieval config, if set
        retrieval_config = request.json.get('retrieval_config', {})
        if not retrieval_config:
            logger.warn('No retrieval config provided, using default retrieval config')

        # add llm model to retrieval config
        if llm_model is not None:
            retrieval_config['llm_model_name'] = llm_model

        # add prompt template to retrieval config
        if prompt_template is not None:
            retrieval_config['rag_prompt_template'] = prompt_template

        # add llm provider to retrieval config if set
        llm_provider = request.json.get('model_provider')
        if llm_provider is not None:
            retrieval_config['llm_provider'] = llm_provider

        # build rag pipeline
        rag_pipeline = table.build_rag_pipeline(retrieval_config)

        # get response from rag pipeline
        rag_response = rag_pipeline(query)
        response = {
            'message': {
                'content': rag_response.get('answer'),
                'context': rag_response.get('context'),
                'role': 'assistant'
            }
        }

        return response



import pandas as pd
from ckanapi import RemoteCKAN
from urllib.parse import urlparse, parse_qs

from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse,
    HandlerStatusResponse,
    RESPONSE_TYPE,
)
from mindsdb_sql.parser import ast
from mindsdb.utilities import log
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

logger = log.getLogger(__name__)


class DatasetsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        limit = query.limit.value if query.limit else 1000
        packages = self.list(conditions, limit)
        return pd.DataFrame(packages)

    def list(self, conditions=None, limit=1000):
        self.handler.connect()
        package_list = self.handler.call_ckan_api("package_search", {"rows": limit})
        packages = package_list.get("results", [])

        data = []
        # Get only datastore active resources
        for pkg in packages:
            datastore_active_resources = [
                r for r in pkg.get("resources", []) if r.get("datastore_active")
            ]
            data.append(
                {
                    "id": pkg.get("id"),
                    "name": pkg.get("name"),
                    "title": pkg.get("title"),
                    "num_resources": len(pkg.get("resources", [])),
                    "num_datastore_active_resources": len(datastore_active_resources),
                }
            )

        return pd.DataFrame(data)

    # Define the columns that will be returned by the table
    # Maybe we can make this dynamic in the future
    def get_columns(self):
        return [
            "id",
            "name",
            "title",
            "num_resources",
            "num_datastore_active_resources",
        ]


class ResourceIDsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        limit = query.limit.value if query.limit else 1000

        resources = self.list(conditions, limit)
        return pd.DataFrame(resources)

    def list(self, conditions=None, limit=1000):
        self.handler.connect()
        package_list = self.handler.call_ckan_api("package_search", {"rows": limit})
        packages = package_list.get("results", [])

        data = []
        for package in packages:
            for resource in package.get("resources", []):
                # Get only datastore active resources
                if resource.get("datastore_active"):
                    data.append(
                        {
                            "id": resource.get("id"),
                            "package_id": package.get("id"),
                            "name": resource.get("name"),
                            "format": resource.get("format"),
                            "url": resource.get("url"),
                            "datastore_active": resource.get("datastore_active"),
                        }
                    )
                if len(data) >= limit:
                    break
            if len(data) >= limit:
                break

        return pd.DataFrame(data)

    def get_columns(self):
        return [
            "id",
            "package_id",
            "name",
            "format",
            "url",
            "datastore_active",
        ]


class DatastoreTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        limit = query.limit.value if query.limit else None

        resource_id = None
        other_conditions = []
        for condition in conditions:
            if isinstance(condition, list) and len(condition) == 3:
                op, col, val = condition
                if col == "resource_id" and op == "=":
                    resource_id = val
                else:
                    other_conditions.append(condition)

        if not resource_id:
            message = "Please provide a resource_id in your query. Example: SELECT * FROM datastore WHERE resource_id = 'your_resource_id'"
            df = pd.DataFrame({"message": [message]})
            return df

        data = self.query_datastore(resource_id, other_conditions, limit)
        return pd.DataFrame(data)

    def query_datastore(self, resource_id, conditions, limit):
        params = {
            "resource_id": resource_id,
            "limit": 1000,  # Start with default limit of 1000 records
        }

        # Add filters based on conditions
        filters = {}
        for condition in conditions:
            op, col, val = condition
            if op == "=":
                filters[col] = val
        if filters:
            params["filters"] = filters

        all_records = []
        while True:
            result = self.handler.call_ckan_api("datastore_search", params)
            records = result.get("records", [])
            all_records.extend(records)

            # Check if we've reached the desired limit
            if limit and len(all_records) >= limit:
                logger.info(f"Reached limit of {limit} records")
                all_records = all_records[:limit]
                break

            # Check for next page
            if "_links" in result and "next" in result["_links"]:
                next_url = result["_links"]["next"]
                parsed_url = urlparse(next_url)
                query_params = parse_qs(parsed_url.query)
                params["offset"] = query_params.get("offset", ["0"])[0]
                logger.info(f"Fetching next page with offset {params['offset']}")
            else:
                logger.info("No more pages to fetch")
                break  # No more pages

        df = pd.DataFrame(all_records)

        # Include metadata about the resource
        metadata = self.handler.call_ckan_api("resource_show", {"id": resource_id})
        df["_resource_id"] = resource_id
        df["_resource_name"] = metadata.get("name")
        df["_resource_format"] = metadata.get("format")
        df["_resource_last_modified"] = metadata.get("last_modified")

        return df

    def get_columns(self):
        return [field["id"] for field in self.fields]


class CkanHandler(APIHandler):
    name = "ckan"

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.connection = None
        self.is_connected = False
        self.connection_args = kwargs.get("connection_data", {})

        self.datasets_table = DatasetsTable(self)
        self.resources_table = ResourceIDsTable(self)
        self.datastore_table = DatastoreTable(self)

        self._register_table("datasets", self.datasets_table)
        self._register_table("resources", self.resources_table)
        self._register_table("datastore", self.datastore_table)

    def connect(self):
        if self.is_connected:
            return self.connection

        url = self.connection_args.get("url")
        api_key = self.connection_args.get("api_key")
        if not url:
            raise ValueError("CKAN URL is required")

        try:
            self.connection = RemoteCKAN(url, apikey=api_key)
            self.is_connected = True
            logger.info(f"Successfully connected to CKAN at {url}")
        except Exception as e:
            logger.error(f"Error connecting to CKAN: {e}")
            raise ConnectionError(f"Failed to connect to CKAN: {e}")

        return self.connection

    def check_connection(self) -> HandlerStatusResponse:
        try:
            self.connect()
            return HandlerStatusResponse(success=True)
        except Exception as e:
            logger.error(f"Error checking connection: {e}")
            return HandlerStatusResponse(success=False, error_message=str(e))

    def call_ckan_api(self, method_name: str, params: dict):
        connection = self.connect()
        method = getattr(connection.action, method_name)

        try:
            result = method(**params)
            return result
        except Exception as e:
            logger.error(f"Error calling CKAN API: {e}")
            raise RuntimeError(f"Failed to call CKAN API: {e}")

    def native_query(self, query: str) -> HandlerResponse:
        method, params = self.parse_native_query(query)
        try:
            result = self.call_ckan_api(method, params)
            if isinstance(result, list):
                df = pd.DataFrame(result)
            elif isinstance(result, dict):
                df = pd.DataFrame([result])
            else:
                df = pd.DataFrame([{"result": result}])
            return HandlerResponse(RESPONSE_TYPE.TABLE, df)
        except Exception as e:
            logger.error(f"Error executing native query: {e}")
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))

    @staticmethod
    def parse_native_query(query: str):
        parts = query.split(":")
        if len(parts) != 2:
            raise ValueError(
                "Invalid query format. Expected 'method_name:param1=value1,param2=value2'"
            )
        method = parts[0].strip()
        params = {}
        if parts[1].strip():
            param_pairs = parts[1].split(",")
            for pair in param_pairs:
                key, value = pair.split("=")
                params[key.strip()] = value.strip()

        return method, params
